/*
 * Copyright (C) Tom Bocklisch <https://github.com/tmbo>
 */
package de.hpi.kdd.mixedfeature.datagen

import java.io.PrintWriter
import java.nio.file.Path

import breeze.io.CSVWriter
import breeze.linalg.DenseMatrix

/**
  * Helper object to write a generated dataset in arff file format. The file
  * can afterwards be read by WEKA or arff libraries.
  */
object ARFFWriter {
  /**
    * Write the passed ds to the folder.
    *
    * @param folder      target folder
    * @param dataSetName name of the dataset
    * @param ds          dataset
    */
  def write(folder: Path, dataSetName: String, ds: GeneratedDataSet): Unit = {
    // Print info to file
    val featureIdxs = (0 until ds.info.props.numberOfFeatures).toList

    new PrintWriter(folder.resolve(dataSetName + ".arff").toFile) {
      println(s"@relation $dataSetName")
      featureIdxs.foreach { idx =>
        val typ = ds.info.nominalFeatures.get(idx) match {
          case Some(nominalProps) => (0 until nominalProps.numberOfValues).map(_.toDouble).mkString("{", ",", "}")
          case _ => "numeric"
        }

        println(s"@attribute V$idx $typ")
      }
      println(s"@attribute class {0.0, 1.0}")
      println("@data")
      val XY: DenseMatrix[Double] = DenseMatrix.horzcat(ds.X, ds.y.asDenseMatrix.t)
      val stringified = IndexedSeq.tabulate(XY.rows, XY.cols)(XY(_, _).toString)
      CSVWriter.write(this, stringified, separator = ',', quote = '\u0000', escape = '\\')
      close()
    }
  }
}
