package de.hpi.kdd.mixedfeature.datagen

import breeze.stats.distributions.{RandBasis, ThreadLocalRandomGenerator}
import org.apache.commons.math3.random.MersenneTwister

import scala.util.Random

/**
  * Properties for the data generation
  *
  * @param numberOfIndependent   number of independent features
  * @param numberOfUseful        number of features from the independent features that are useful for prediction
  * @param numberOfDependent     number of features that are linear combinations of the independent features
  * @param numberOfRepeated      number of duplicate features
  * @param numberOfInstances     number of instances in the dataset
  * @param corrClusters          correlation clusters
  * @param nominalFeatures       nominal features
  * @param inherentYError        underlying error in generated labels
  * @param shouldShiftAndScale   indicator if the feature values should be denormalized
  * @param shouldShuffleFeatures indicator if the features should be shuffeled
  * @param seed                  seed to reproduce the dataset
  */
case class DataGenProperties(
                              numberOfIndependent: Int,
                              numberOfUseful: Int,
                              numberOfDependent: Int,
                              numberOfRepeated: Int,
                              numberOfInstances: Int,
                              corrClusters: List[CorrClusterProperties],
                              nominalFeatures: List[NominalFeatureProperties],
                              inherentYError: Double,
                              shouldShiftAndScale: Boolean,
                              shouldShuffleFeatures: Boolean,
                              seed: Int = 0
                            ) {
  // Total number of features
  def numberOfFeatures = numberOfDependent + numberOfIndependent + numberOfRepeated

  // Total number of features in a cluster
  def numberOfClusterFeatures = corrClusters.map(_.size).sum

  // Total number of nominal features
  def numberOfNominal = nominalFeatures.length

  // Number of clusters
  def numClusters = corrClusters.length

  // random number generator
  @transient
  lazy val rand = new RandBasis(new ThreadLocalRandomGenerator(new MersenneTwister(seed)))
}

/**
  * Helper to generate & validate data properties
  */
object DataGenProperties {
  /**
    * Validation for data generation properties
    * @param props properties
    * @return if the properties are valid
    */
  def isProperConfiguration(props: DataGenProperties): Boolean = {
    // we can not have more useful than independent features
    props.numberOfUseful <= props.numberOfIndependent &&
      // we can at max convert all features into nominal ones
      props.numberOfNominal <= props.numberOfFeatures &&
      // a nominal feature needs to have at least two states
      props.nominalFeatures.forall(_.numberOfValues >= 2) &&
      // the underlying error probability can't exceed one and must be positive
      props.inherentYError < 1 && props.inherentYError >= 0
  }

  def init(numberOfIndependent: Int,
           numberOfUseful: Int,
           numberOfDependent: Int,
           numberOfRepeated: Int,
           numberOfInstances: Int,
           corrClusters: List[CorrClusterProperties],
           nominalFeatures: List[NominalFeatureProperties],
           inherentYError: Double,
           shouldShiftAndScale: Boolean,
           shouldShuffleFeatures: Boolean,
           seed: Int = Random.nextInt): Option[DataGenProperties] = {

    val props = DataGenProperties(
      numberOfIndependent,
      numberOfUseful,
      numberOfDependent,
      numberOfRepeated,
      numberOfInstances,
      corrClusters,
      nominalFeatures,
      inherentYError,
      shouldShiftAndScale,
      shouldShuffleFeatures,
      seed
    )
    if (isProperConfiguration(props))
      Some(props)
    else
      None
  }
}
