package de.hpi.kdd.mixedfeature.datagen

import breeze.linalg.{DenseMatrix, _}
import breeze.numerics.{abs, exp, log, signum}
import breeze.stats.distributions.{RandBasis, ThreadLocalRandomGenerator}
import com.typesafe.scalalogging.LazyLogging
import de.hpi.kdd.mixedfeature.util.Extensions._
import de.hpi.kdd.mixedfeature.util.{ColoredText, Direction}
import org.apache.commons.math3.random.MersenneTwister

/**
  * Helper object to generate artificial datasat according to the NIPS process
  */
object DataGenerator extends LazyLogging {

  /**
    * Generate artificial data using the provided properties
    *
    * @param prop properties of the dataset to be generated
    * @return generated dataset
    */
  def generate(prop: DataGenProperties): GeneratedDataSet = {

    // Draw n features independently at random
    val indepX = DenseMatrix.rand(prop.numberOfInstances, prop.numberOfIndependent, gaussian(0, 1, prop))
    var X = indepX.copy

    // Each additional feature defined as a linear combination of the previous ones
    // Provided by a matrix of coefficients of dimention (num_indep_feat, num_dep_feat)
    if (prop.numberOfDependent > 0) {
      var dependentFactors: DenseMatrix[Double] = DenseMatrix.rand(prop.numberOfIndependent, prop.numberOfDependent, prop.rand.uniform)
      dependentFactors *= 2.0
      dependentFactors -= 1.0
      val depX = indepX * dependentFactors
      X = DenseMatrix.horzcat(X, depX)
    }

    // Repeat num_repeat_feat features, chosen at random among indep and dep feat
    val numFeatureThatCanBeDuplicated = prop.numberOfIndependent + prop.numberOfDependent
    val duplicatedFeatures = DenseMatrix
      .rand(prop.numberOfRepeated, 1, prop.rand.randInt(numFeatureThatCanBeDuplicated))
      .toDenseVector
      .toScalaVector()

    if (prop.numberOfRepeated > 0) {
      val dupX = X(::, duplicatedFeatures)
      X = DenseMatrix.horzcat(X, dupX.toDenseMatrix)
    }

    // Add a little bit of random noise
    val noise = DenseMatrix.rand(prop.numberOfInstances, prop.numberOfFeatures, gaussian(0, 0.1, prop))
    X = X + noise

    val (y, featureWeights) = generateLabels(X, prop)

    val (xWithNominals, nominalFeatures) = generateNominal(X, prop)

    X = xWithNominals

    // Randomly shift and scale
    // TODO: fix, doesn't work with nominal features
//    if (prop.shouldShiftAndScale) {
//      val shift = DenseMatrix.rand(1, prop.numberOfFeatures, prop.rand.uniform)
//      val scale = DenseMatrix.rand(1, prop.numberOfFeatures, prop.rand.gaussian(1, 1000))
//      X = X + shift.repmat(prop.numberOfInstances, Direction.vertical)
//      X = X :* scale.repmat(prop.numberOfInstances, Direction.vertical)
//    }

    // randomly permute the features
    val fidx =
      if (prop.shouldShuffleFeatures) {
        val rp = prop.rand.permutation(prop.numberOfFeatures).draw()
        val mapping = (0 until prop.numberOfFeatures).zip(rp).toMap
        val XBuf = X.copy
        mapping.foreach {
          case (i, shuffledIdx) =>
            X(::, shuffledIdx) := XBuf(::, i)
        }

        mapping
      } else
        (0 until prop.numberOfFeatures).zip(0 until prop.numberOfFeatures).toMap

    val usefulFeatureList = (0 until prop.numberOfUseful).map(i => fidx(i))
    val dependentFeatureList = (prop.numberOfIndependent until (prop.numberOfIndependent + prop.numberOfDependent)).map(i => fidx(i))
    val duplicatedFeatureList = duplicatedFeatures.map(i => fidx(i)).zip(((prop.numberOfIndependent + prop.numberOfDependent) until prop.numberOfFeatures).map(i => fidx(i)))
    val (_, clusterList) = prop.corrClusters.map(_.size).foldLeft((0, Array.empty[(List[Int], Double)])) {
      case ((offset, clusters), clusterSize) =>
        val w = featureWeights.valueAt(prop.numberOfUseful + clusters.length)
        val clusterInfo = (offset until (clusterSize + offset)).toList.map(i => fidx(i)) -> w
        (offset + clusterSize, clusters :+ clusterInfo)
    }

    val nominalFeatureList = nominalFeatures.map(i => fidx(i._1) -> i._2).toMap
    val classDist = List(sum(y).toInt, prop.numberOfInstances - sum(y).toInt)

    val weights = DenseVector.zeros[Double](prop.numberOfFeatures)
    (0 until prop.numberOfUseful).map(i => fidx(i)).zipWithIndex.foreach {
      case (shuffeledIdx, i) =>
        weights.update(shuffeledIdx, featureWeights.valueAt(i))
    }

    GeneratedDataSet(X, y,
      GeneratedDataSetInfo(prop, usefulFeatureList, dependentFeatureList, duplicatedFeatureList, nominalFeatureList, clusterList, weights, classDist))
  }

  def createClusterVectors(X: DenseMatrix[Double], prop: DataGenProperties) = {
    var clusterVectors = scala.collection.immutable.Vector.empty[DenseVector[Double]]

    prop.corrClusters.map(_.size).foldLeft(0) {
      case (startIdx, currentClusterSize) =>
        val c = X(::, startIdx until startIdx + currentClusterSize)

        val signC = (signum(c) + 1.0) / 2.0
        val XORSign: DenseVector[Double] = (sum(signC(*, ::)) :%= 2.0) * 2.0 - 1.0

        val logified = log(abs(c))
        val XOR = exp(sum(logified(*, ::)) * (1.0 / currentClusterSize)) :* XORSign

        clusterVectors = clusterVectors :+ XOR
        startIdx + currentClusterSize
    }

    DenseVector.horzcat(clusterVectors: _*)
  }

  private def generateNominal(X: DenseMatrix[Double], prop: DataGenProperties): (DenseMatrix[Double], List[(Int, NominalFeatureProperties)]) = {
    val shuffledIndices = prop.rand.permutation(prop.numberOfFeatures).draw().toList

    prop.nominalFeatures.zipWithIndex.foreach {
      case (nominalProps, idx) =>
        val diced = prop.rand.permutation(nominalProps.numberOfValues).draw().toArray
        val (v1, v2) = diced.splitAt(nominalProps.numberOfValues / 2)
        val i = shuffledIndices(idx)

        val nominalValues = X(::, i).map { c =>
          if (c >= 0.0)
            randomElement(v1, prop).toDouble
          else
            randomElement(v2, prop).toDouble
        }
        X(::, i) := nominalValues
    }

    (X, shuffledIndices.take(prop.numberOfNominal).zip(prop.nominalFeatures))
  }

  private def generateLabels(X: DenseMatrix[Double], prop: DataGenProperties) = {
    var predX = X(::, 0 until prop.numberOfUseful)
    // Matrix of 1's and 0's for prediction
    if (prop.numberOfClusterFeatures > 0) {
      val clusters = createClusterVectors(X, prop)

      // Features used for prediction
      predX = DenseMatrix.horzcat(predX, clusters)
    }

    val numClusterFeatures = prop.numberOfClusterFeatures
    val numClusters = prop.numClusters

    // A clamp factor limits the influence of a feature up to the factor. Hence if the factor is 0.3 a feature
    // can have a maximum correlation of 0.3. This is especially used for features that are part of a cluster.
    val clusterClampFactors = prop.corrClusters.flatMap(c => Array.fill(c.size)(c.maxIndividualFeatureCorr)).toArray
    val remainingFeatureClampFactors = Array.fill(prop.numberOfUseful - numClusterFeatures + numClusters)(0.9)
    val clampFactor = DenseVector(Array.concat(clusterClampFactors, remainingFeatureClampFactors))

    val numWeights = prop.numberOfUseful + numClusters

    // Label calculation using a linear combination of the useful features
    val w = DenseVector.rand(numWeights, prop.rand.uniform) :* clampFactor
    var y: DenseVector[Double] = signum(predX * w)

    // Add random y label errors
    val numberWronglyLabeled = math.round(prop.numberOfInstances * prop.inherentYError).toInt
    val fi = nextRand(prop).permutation(prop.numberOfInstances).draw().take(numberWronglyLabeled)
    y(fi) *= -1.0

    y = (y + 1.0) / 2.0
    (y, w)
  }

  /**
    * Choose a random element from the array
    */
  private def randomElement[A](a: Array[A], prop: DataGenProperties): A = {
    prop.rand.choose(a).draw()
  }

  /**
    * Generate the next random number generator
    */
  private def nextRand(prop: DataGenProperties) = {
    val seed = prop.rand.randInt.draw()
    new RandBasis(new ThreadLocalRandomGenerator(new MersenneTwister(seed)))
  }

  /**
    * Next gaussian using the random number generator from props
    */
  private def gaussian(m: Double, s: Double, prop: DataGenProperties) = {
    nextRand(prop).gaussian(m, s)
  }

}
