package de.hpi.kdd.mixedfeature.datagen

import breeze.linalg.{DenseMatrix, DenseVector, _}
import de.hpi.kdd.mixedfeature.util.ColoredText

/**
  * Collected information about a generated dataset
  *
  * @param X feature matrix
  * @param y label vector
  */
@SerialVersionUID(100L)
case class GeneratedDataSet(
                             X: DenseMatrix[Double],
                             y: DenseVector[Double],
                             info: GeneratedDataSetInfo)


/**
  * Collected information about a generated dataset
  *
  * @param props              properties used for generation
  * @param usefulFeatures     indices of the useful features
  * @param dependentFeatures  indices of the dependent features
  * @param duplicatedFeatures indices of the duplicated features
  * @param nominalFeatures    indices of the nominal features
  * @param clusterInfo        information about generated clusters
  * @param weights            importance of the different useful features
  * @param classDist          class label distribution
  */
@SerialVersionUID(101L)
case class GeneratedDataSetInfo(
                                 props: DataGenProperties,
                                 usefulFeatures: Seq[Int],
                                 dependentFeatures: Seq[Int],
                                 duplicatedFeatures: Seq[(Int, Int)],
                                 nominalFeatures: Map[Int, NominalFeatureProperties],
                                 clusterInfo: Array[(List[Int], Double)],
                                 weights: DenseVector[Double],
                                 classDist: List[Int]
                               ) extends ColoredText {
  lazy val clusters = clusterInfo.map(_._1)

  lazy val nominalFeatureIndices = nominalFeatures.keySet

  lazy val totalWeights = sum(weights) + clusterInfo.map(_._2).sum

  override def toString: String = {
    s"GeneratedDataSet(\n" +
      s"\tuseful=${usefulFeatures.sorted}\n" +
      s"\tdependent=${dependentFeatures.sorted}\n" +
      s"\tduplicated=${duplicatedFeatures.sortBy(_._1)}\n" +
      s"\tnominal=${nominalFeatures.toList.sortBy(_._1)}\n" +
      s"\tclusters=${clusters.map(_.sorted.mkString("(", ", ", ")")).mkString("[", ", ", "]")}\n" +
      s"\tdistribution=$classDist\n" +
      s"\tseed=${props.seed}\n" +
      s")"
  }

  def printFeatureInfo = {
    (0 until props.numberOfFeatures).map { fidx =>
      f"V$fidx%2d ${clusterOf(fidx)} ${relevancyOf(fidx)}%s ${weights(fidx)}%.2f ${clusterWeightOf(fidx)}%s ${featureTypeOf(fidx)}%s"
    }.mkString("\n")
  }

  private def relevancyOf(idx: Int) = {
    if (usefulFeatures.contains(idx))
      inColor("green", "rel")
    else if (dependentFeatures.contains(idx))
      inColor("yellow", "dep")
    else if (duplicatedFeatures.exists(_._2 == idx))
      inColor("red", "dup")
    else
      inColor("white", "irr")
  }

  private def featureTypeOf(idx: Int) = nominalFeatures.get(idx) match {
    case Some(nominalProps) => (0 until nominalProps.numberOfValues).map(_.toDouble).mkString("{", ",", "}")
    case _ => "numeric"
  }

  private def clusterId(idx: Int) = {
    val id = clusters.indexWhere(_.contains(idx))
    if (id == -1)
      None
    else
      Some(id)
  }

  private def clusterWeightOf(idx: Int) = {
    clusterId(idx) match {
      case Some(cluster) =>
        f"${clusterInfo(cluster)._2}%.2f"
      case _ =>
        "    "
    }
  }

  private def clusterOf(idx: Int) = {
    clusterId(idx) match {
      case Some(cluster) =>
        inColor(cluster, cluster.toString)
      case _ =>
        " "
    }
  }
}